/* db_packet.c - packet transfer routines
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "db_packet.h"
#include "db_crctab.h"
#include "../runtime/db_vm.h"

/* packet format:

  SOH pkt# type length hdrchk length*data crc1 crc2

*/

#define RCV_SOH		1
#define RCV_PKTN	2
#define RCV_TYPE	3
#define RCV_LEN		4
#define RCV_CHK		5
#define RCV_DATA	6
#define RCV_CRC1	7
#define RCV_CRC2	8

/* packet header and crc lengths */
#define PKTHDRLEN   5
#define PKTCRCLEN   2

/* maximum length of a frame */
#define FRAMELEN    (PKTHDRLEN + PKTMAXLEN + PKTCRCLEN)

typedef struct {
    int state;
	int timeout;
	int lastCharTime;
    uint8_t rcv_nextPktn;
    uint8_t rcv_pktn;
    uint8_t rcv_type;
    uint8_t rcv_len;
    uint8_t rcv_chk;
    uint8_t rcv_cnt;
    uint8_t rcv_packet[PKTMAXLEN];
    uint16_t rcv_crc;
    uint8_t snd_pktn;
    uint16_t load_coff;
	uint16_t load_ccnt;
    uint16_t load_ioff;
	uint16_t load_icnt;
    VMFILE *fp;
} PktState;

#define INIT_HDR1   1
#define INIT_HDR2   2
#define INIT_DATA   3

/* protocol characters */
#define SOH     0x01    /* start of a packet */

#define RETRY   5       /* number of times to retry */
#define XTIME   1       /* timeout waiting ACK/NAK */

static int portInit = VMFALSE;
static PktState pkt;

static void BuildFrame(int type, uint8_t *hdr, uint8_t *crc, uint8_t *buf, int len);
static void SendCurrentFrame(uint8_t *hdr, uint8_t *crc, uint8_t *buf, int len);
static void Init(void);
static void Purge(void);
static int CheckHeader(void);
static int CheckPacket(int *pAckNak);
static int HandlePacket(void);
static void ProcessDebug(void);
static void ProcessGetLine(void);
static void ProcessLine(void);
static void ProcessOpen(void);
static void ProcessLoad(void);
static void ProcessRun(void);
static void ProcessData(void);
static void ProcessClose(void);
static void ProcessLoadC(void);
static void ProcessLoadI(void);
static void ProcessStart(void);

int PktOpen(char *port)
{
    /* initialize the serial port */
    if (!portInit) {
        if (!md_init(port))
            return VMFALSE;
        portInit = VMTRUE;
    }

    /* initialize the packet handler */
    Init();

    /* no file transfer currently in progress */
    pkt.fp = NULL;

    /* return successfully */
    return VMTRUE;
}

void PktClose(void)
{
    if (portInit) {
        portInit = VMFALSE;
        md_done();
    }
}

/* PktStartSequence - send a packet and wait for an ACK/NAK */
void PktStartSequence(void)
{
    pkt.snd_pktn = 1;
}

/* PktSendPacket - send a packet and wait for an ACK/NAK */
int PktSendPacket(int type, uint8_t *buf, int len)
{
    uint8_t hdr[PKTHDRLEN];
    uint8_t crc[PKTCRCLEN];
    int retries;

    /* build the packet to send */
    BuildFrame(type, hdr, crc, buf, len);

    /* send data and wait for an ACK */
    for (retries = RETRY; retries > 0; --retries) {

        /* send the current frame */
        SendCurrentFrame(hdr, crc, buf, len);

        /* wait for ACK */
        if (PktPoll(XTIME) == PKT_ACK)
            return VMTRUE;
    }

    /* retry count ran out */
    return VMFALSE;
}

void PktSendFrame(int type, uint8_t *buf, int len)
{
    uint8_t hdr[PKTHDRLEN];
    uint8_t crc[PKTCRCLEN];
    BuildFrame(type, hdr, crc, buf, len);
    SendCurrentFrame(hdr, crc, buf, len);
}

static void BuildFrame(int type, uint8_t *hdr, uint8_t *crc, uint8_t *buf, int len)
{
    uint16_t crc16 = 0;

    /* setup the frame header */
    hdr[0] = SOH;                       /* SOH */
    hdr[1] = pkt.snd_pktn;              /* packet# */
    hdr[2] = type;                      /* type type */
    hdr[3] = len;                       /* data length */
    hdr[4] = hdr[1] + hdr[2] + hdr[3];  /* header checksum */

    /* compute the crc */
    while (--len >= 0)
        crc16 = updcrc(crc16, *buf++);
    crc16 = updcrc(crc16, '\0');
    crc16 = updcrc(crc16, '\0');

    /* add the crc to the frame */
    crc[0] = (uint8_t)(crc16 >> 8);
    crc[1] = (uint8_t)crc16;

    /* increment the packet number */
    if (type >= PKT_SEQ)
        ++pkt.snd_pktn;
}

static void SendCurrentFrame(uint8_t *hdr, uint8_t *crc, uint8_t *buf, int len)
{
    md_write(hdr, PKTHDRLEN);
    md_write(buf, len);
    md_write(crc, PKTCRCLEN);
}

int PktPoll(int timeout)
{
    int ch;

    /* handle received characters */
    if ((ch = md_get(timeout)) == MD_TIMEOUT) {
		// handle timeout
	}
	else {
        do {
            switch (pkt.state) {
            case RCV_SOH:
                if (ch == SOH)
                    pkt.state = RCV_PKTN;
				else
					Purge();
                break;
            case RCV_PKTN:
                pkt.rcv_pktn = ch;
                pkt.state = RCV_TYPE;
                break;
            case RCV_TYPE:
                pkt.rcv_type = ch;
                pkt.state = RCV_LEN;
                break;
            case RCV_LEN:
                pkt.rcv_len = ch;
                pkt.state = RCV_CHK;
                break;
             case RCV_CHK:
                pkt.rcv_chk = ch;
                if (CheckHeader()) {
                    pkt.rcv_crc = 0;
                    if (pkt.rcv_len == 0)
                        pkt.state = RCV_CRC1;
                    else {
                        pkt.rcv_cnt = 0;
                        pkt.state = RCV_DATA;
                    }
                }
                else {
					Purge();
					pkt.state = RCV_SOH;
                }
                break;
           case RCV_DATA:
                pkt.rcv_packet[pkt.rcv_cnt] = ch;
                pkt.rcv_crc = updcrc(pkt.rcv_crc, ch);
                if (++pkt.rcv_cnt >= pkt.rcv_len)
                    pkt.state = RCV_CRC1;
                break;
            case RCV_CRC1:
                pkt.rcv_crc = updcrc(pkt.rcv_crc, ch);
                pkt.state = RCV_CRC2;
                break;
            case RCV_CRC2:
                pkt.rcv_crc = updcrc(pkt.rcv_crc, ch);
                pkt.state = RCV_SOH;
                HandlePacket();
                switch (pkt.rcv_type) {
                case PKT_ACK:
                case PKT_NAK:
                    return pkt.rcv_type;
                }
                break;
            }
        } while ((ch = md_get(timeout)) != MD_TIMEOUT);
    }

    return PKT_NONE;
}

static void Init(void)
{
    memset(&pkt, 0, sizeof(pkt));
    pkt.state = RCV_SOH;
    pkt.snd_pktn = 1;
    pkt.rcv_nextPktn = 1;
}

static void Purge(void)
{
	while (md_get(0) != MD_TIMEOUT)
		;
}

static int CheckHeader(void)
{
    /* check the header checksum */
    if (pkt.rcv_pktn + pkt.rcv_type + pkt.rcv_len != pkt.rcv_chk)
        return VMFALSE;

    /* header is okay */
    return VMTRUE;
}

static int CheckPacket(int *pAckNak)
{
    /* check the crc and packet length */
    if (pkt.rcv_crc != 0 || pkt.rcv_len > PKTMAXLEN) {
        *pAckNak = pkt.rcv_type < PKT_BASE ? PKT_NONE : PKT_NAK;
        return VMFALSE;
    }

    /* handle protocol packets */
    if (pkt.rcv_type < PKT_BASE)
        *pAckNak = PKT_NONE;

    /* handle non-protocol packets */
    else if (pkt.rcv_type < PKT_SEQ)
        *pAckNak = PKT_ACK;
        
    /* handle non-protocol sequence packets */
    else {
        
        /* check for a duplicate packets */
        if (pkt.rcv_pktn == ((pkt.rcv_nextPktn - 1) & 0xff)) {
            *pAckNak = PKT_ACK;
            return VMFALSE;
        }

        /* check for a packets sequence error */
        else if (pkt.rcv_pktn != pkt.rcv_nextPktn) {
            *pAckNak = PKT_NAK;
            return VMFALSE;
        }

        /* update the next expected packet number */
        ++pkt.rcv_nextPktn;

        /* acknowledge the packet */
        *pAckNak = PKT_ACK;
    }

    /* handle good packets */
    return VMTRUE;
}

static int HandlePacket(void)
{
    int handle, ackNak;

    /* check the packet */
    handle = CheckPacket(&ackNak);

    /* handle the packet */
    if (handle) {
        switch (pkt.rcv_type) {
        case PKT_ACK:
            break;
        case PKT_NAK:
            break;
        default:
            switch (pkt.rcv_type) {
            case PKT_DEBUG:
                ProcessDebug();
                break;
            case PKT_GETLINE:
                ProcessGetLine();
                break;
            case PKT_LINE:
                ProcessLine();
                break;
            case PKT_OPEN:  // OPEN DATA* CLOSE
                pkt.rcv_nextPktn = 1;
                ProcessOpen();
                break;
            case PKT_LOAD:  // LOAD LOADC* LOADI* START
                pkt.rcv_nextPktn = 1;
                ProcessLoad();
                break;
            case PKT_RUN:
                ProcessRun();
                break;
            case PKT_DATA:
                ProcessData();
                break;
            case PKT_CLOSE:
                ProcessClose();
                break;
            case PKT_LOADC:
                ProcessLoadC();
                break;
            case PKT_LOADI:
                ProcessLoadI();
                break;
            case PKT_START:
                ProcessStart();
                break;
            default:
                 break;
            }
            break;
        }  
    }

    /* send the acknowledgement */
    if (ackNak != PKT_NONE)
        PktSendFrame(ackNak, NULL, 0);

    /* packet was processed */
    return VMTRUE;
}

static void ProcessDebug(void)
{
#ifdef WIN32
    char *str;
    
    /* get the debug string */
    str = (char *)pkt.rcv_packet;

    /* display it */
    VM_writedebug(pkt.rcv_packet, pkt.rcv_len);
#endif
}

static void ProcessGetLine(void)
{
#ifdef WIN32
    VM_getline();
#endif
}

static void ProcessLine(void)
{
#ifndef WIN32
    SetInputLine(pkt.rcv_packet, pkt.rcv_len);
#endif
}

static void ProcessOpen(void)
{
#ifndef WIN32
    char *name;
    
    /* close any currently open file */
    ProcessClose();
    
    /* get the filename */
    name = (char *)pkt.rcv_packet;

    /* check for the start of a new transfer */
    if (!(pkt.fp = VM_fopen(name, "wb")))
        VM_printf("Open failed: %s\n", name);
#endif
}

static void ProcessData(void)
{
#ifndef WIN32
    if (pkt.fp)
        VM_fwrite(pkt.rcv_packet, pkt.rcv_len, pkt.fp);
#endif
}

static void ProcessClose(void)
{
#ifndef WIN32
    if (pkt.fp) {
        VM_fclose(pkt.fp);
        pkt.fp = NULL;
    }
#endif
}

static void ProcessRun(void)
{
#ifndef WIN32
    char *name;
    
    /* get the filename */
    name = (char *)pkt.rcv_packet;

    /* load a new program */
    if (!LoadImageFile(&interp, name))
        VM_printf("Run failed: %s\n", name);
#endif
}

static void ProcessLoad(void)
{
#ifndef WIN32
	FileHdr *hdr = (FileHdr *)pkt.rcv_packet;
	if (pkt.rcv_len == sizeof(FileHdr) && StartImageLoad(&interp, hdr)) {
        pkt.load_coff = 0;
        pkt.load_ccnt = hdr->codeSize;
        pkt.load_ioff = 0;
        pkt.load_icnt = hdr->initSize;
    }
    else
        VM_printf("Load failed\n");
#endif
}

static void ProcessLoadC(void)
{
#ifndef WIN32
	if (pkt.load_ccnt > 0) {
		int cnt;

		/* handle a short last packet */
        if ((cnt = pkt.rcv_len) > pkt.load_ccnt)
			cnt = pkt.load_ccnt;

		/* load the code */
        VM_LoadCode(pkt.load_coff, pkt.rcv_packet, cnt);

		/* update the pointers */
        pkt.load_coff += cnt;
		pkt.load_ccnt -= cnt;
	}
#endif
}

static void ProcessLoadI(void)
{
#ifndef WIN32
	if (pkt.load_icnt > 0) {
		int cnt;

		/* handle a short last packet */
        if ((cnt = pkt.rcv_len) > pkt.load_icnt)
			cnt = pkt.load_icnt;

        /* load the initializers */
		VM_LoadInit(pkt.load_ioff, pkt.rcv_packet, cnt);

        /* update the pointers */
		pkt.load_ioff += cnt;
		pkt.load_icnt -= cnt;

        /* process the initialization data */
        ProcessInitData(&interp, pkt.rcv_packet, cnt);
    }
#endif
}

static void ProcessStart(void)
{
#ifndef WIN32
	if (pkt.load_ccnt == 0 && pkt.load_icnt == 0) {
        VM_StoreMemoryImageHeader(&interp);
        FinishImageLoad(&interp);
    }
#endif
}

