/* db_bcode.c - bytecode handling functions
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include "db_compiler.h"

/* local function prototypes */
static void code_expr(ParseContext *c, ParseTreeNode *expr, PVAL *pv);
static void code_shortcircuit(ParseContext *c, int op, ParseTreeNode *expr, PVAL *pv);
static void code_arrayref(ParseContext *c, ParseTreeNode *expr, PVAL *pv);
static void code_call(ParseContext *c, ParseTreeNode *expr, PVAL *pv);
static int GREF_inst(Type *type);
static int GSET_inst(Type *type);
static int AREF_inst(Type *type);
static int ASET_inst(Type *type);
static int LREF_inst(Type *type);
static int LSET_inst(Type *type);
static int AREF1_inst(Type *type);
static int AREF2_inst(Type *type);
static int ASET1_inst(Type *type);
static int ASET2_inst(Type *type);

/* code_lvalue - generate code for an l-value expression */
void code_lvalue(ParseContext *c, ParseTreeNode *expr, PVAL *pv)
{
    code_expr(c, expr, pv);
    chklvalue(c, pv);
}

/* code_rvalue - generate code for an r-value expression */
void code_rvalue(ParseContext *c, ParseTreeNode *expr)
{
    PVAL pv;
    code_expr(c, expr, &pv);
    rvalue(c, &pv);
}

/* code_expr - generate code for an expression parse tree */
static void code_expr(ParseContext *c, ParseTreeNode *expr, PVAL *pv)
{
    pv->type = NodeSrcType(c, expr);
    switch (expr->nodeType) {
    case NodeTypeSymbolRef:
        pv->fcn = expr->symbolRef.fcn;
        pv->val = expr->symbolRef.offset;
        break;
    case NodeTypeStringLit:
        putcbyte(c, OP_LIT);
        putcword(c, AddStringRef(expr->stringLit.string, codeaddr(c)));
        pv->fcn = NULL;
        break;
    case NodeTypeIntegerLit:
        putcbyte(c, OP_LIT);
        putcword(c, expr->integerLit.value);
        pv->fcn = NULL;
        break;
    case NodeTypeFunctionLit:
        putcbyte(c, OP_LIT);
        putcword(c, expr->functionLit.offset);
        pv->fcn = NULL;
        break;
    case NodeTypeUnaryOp:
        code_rvalue(c, expr->unaryOp.expr);
        putcbyte(c, expr->unaryOp.op);
        pv->fcn = NULL;
        break;
    case NodeTypeBinaryOp:
        code_rvalue(c, expr->binaryOp.left);
        code_rvalue(c, expr->binaryOp.right);
        putcbyte(c, expr->binaryOp.op);
        pv->fcn = NULL;
        break;
    case NodeTypeArrayRef:
        code_arrayref(c, expr, pv);
        break;
    case NodeTypeFunctionCall:
        code_call(c, expr, pv);
        break;
    case NodeTypeDisjunction:
        code_shortcircuit(c, OP_BRTSC, expr, pv);
        break;
    case NodeTypeConjunction:
        code_shortcircuit(c, OP_BRFSC, expr, pv);
        break;
    }
}

/* code_shortcircuit - generate code for a conjunction or disjunction of boolean expressions */
static void code_shortcircuit(ParseContext *c, int op, ParseTreeNode *expr, PVAL *pv)
{
    ExprListEntry *entry = expr->exprList.exprs;
    int end = 0;

    code_rvalue(c, entry->expr);
    entry = entry->next;

    do {
        putcbyte(c, op);
        end = putcword(c, end);
        code_rvalue(c, entry->expr);
    } while ((entry = entry->next) != NULL);

    fixupbranch(c, end, codeaddr(c));

    pv->fcn = NULL;
}

/* code_arrayref - code an array reference */
static void code_arrayref(ParseContext *c, ParseTreeNode *expr, PVAL *pv)
{
    /* code the array */
    code_rvalue(c, expr->arrayRef.array);
    
    /* code the first index */
    code_rvalue(c, expr->arrayRef.index1);
    
    /* check for a second index */
    if (expr->arrayRef.index2) {
        code_rvalue(c, expr->arrayRef.index2);
        pv->val = 2;
    }
    
    /* only one index */
    else
        pv->val = 1;
    
    /* setup the element type */
    pv->fcn = code_index;
}

/* code_call - code a function call */
static void code_call(ParseContext *c, ParseTreeNode *expr, PVAL *pv)
{
    ExprListEntry *arg;
    
    /* get the value of the function */
    code_rvalue(c, expr->functionCall.fcn);

    /* code each argument expression */
    for (arg = expr->functionCall.args; arg != NULL; arg = arg->next)
        code_rvalue(c, arg->expr);

    /* call the function */
    putcbyte(c, OP_CALL);
    putcbyte(c, expr->functionCall.argc);

    /* we've got an rvalue now */
    pv->fcn = NULL;
}

/* rvalue - get the rvalue of a partial expression */
void rvalue(ParseContext *c, PVAL *pv)
{
    if (pv->fcn) {
        (*pv->fcn)(c, PV_LOAD, pv);
        pv->fcn = NULL;
    }
}

/* chklvalue - make sure we've got an lvalue */
void chklvalue(ParseContext *c, PVAL *pv)
{
    if (pv->fcn == NULL)
        ParseError(c,"Expecting an lvalue");
}

/* code_literal - compile a literal reference */
void code_literal(ParseContext *c, PValOp fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        switch (pv->type->id) {
        case TYPE_STRING:
            putcbyte(c, OP_LIT);
            putcword(c, AddStringRef(pv->str, codeaddr(c)));
            break;
        default:
            putcbyte(c, OP_LIT);
            putcword(c, pv->val);
        }
        break;
    case PV_STORE:
        ParseError(c, "can't store into a literal");
        break;
    }
}

/* code_global - compile a global variable reference */
void code_global(ParseContext *c, PValOp fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        putcbyte(c, GREF_inst(pv->type));
        putcword(c, pv->val);
        break;
    case PV_STORE:
        putcbyte(c, GSET_inst(pv->type));
        putcword(c, pv->val);
        break;
    }
}

/* GREF_inst - select a global reference instruction */
static int GREF_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
        inst = OP_GREF;
        break;
    case TYPE_BYTE:
        inst = OP_GREFB;
        break;
    case TYPE_ARRAY:
        inst = OP_LIT;
        break;
    case TYPE_STRING:
        inst = OP_GREFS;
        break;
    default:
        Fatal("internal error: GREF_inst");
        break;
    }
    return inst;
}

/* GSET_inst - select a global set instruction */
static int GSET_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_GSET;
        break;
    case TYPE_BYTE:
        inst = OP_GSETB;
        break;
    case TYPE_STRING:
        inst = OP_GSETS;
        break;
    default:
        Fatal("internal error: GSET_inst");
        break;
    }
    return inst;
}

/* code_argument - compile an argument reference */
void code_argument(ParseContext *c, PValOp fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        putcbyte(c, AREF_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    case PV_STORE:
        putcbyte(c, ASET_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    }
}

/* AREF_inst - select an argument reference instruction */
static int AREF_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_LREF;
        break;
    case TYPE_BYTE:
        inst = OP_LREFB;
        break;
    case TYPE_STRING:
        inst = OP_LREFS;
        break;
    default:
        Fatal("internal error: AREF_inst");
        break;
    }
    return inst;
}

/* ASET_inst - select an argument set instruction */
static int ASET_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_LSET;
        break;
    case TYPE_BYTE:
        inst = OP_LSETB;
        break;
    case TYPE_STRING:
        inst = OP_LSETS;
        break;
    default:
        Fatal("internal error: ASET_inst");
        break;
    }
    return inst;
}

/* code_local - compile an local reference */
void code_local(ParseContext *c, PValOp fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        putcbyte(c, LREF_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    case PV_STORE:
        putcbyte(c, LSET_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    }
}

/* LREF_inst - select a local reference instruction */
static int LREF_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
        inst = OP_LREF;
        break;
    case TYPE_BYTE:
        inst = OP_LREFB;
        break;
    case TYPE_ARRAY:
        inst = OP_LOFF;
        break;
    case TYPE_STRING:
        inst = OP_LREF;
        // adjust refCnt
        break;
    default:
        Fatal("internal error: LREF_inst");
        break;
    }
    return inst;
}

/* LSET_inst - select a local set instruction */
static int LSET_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_LSET;
        break;
    case TYPE_BYTE:
        inst = OP_LSETB;
        break;
    case TYPE_STRING:
        inst = OP_LSET;
        // adjust refCnt
        break;
    default:
        Fatal("internal error: LSET_inst");
        break;
    }
    return inst;
}

/* code_index - compile an array reference */
void code_index(ParseContext *c, PValOp fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        if (pv->val == 1)
            putcbyte(c, AREF1_inst(pv->type));
        else
            putcbyte(c, AREF2_inst(pv->type));
        break;
    case PV_STORE:
        if (pv->val == 1)
            putcbyte(c, ASET1_inst(pv->type));
        else
            putcbyte(c, ASET2_inst(pv->type));
        break;
    }
}

/* AREF1_inst - select a 1 index reference instruction */
static int AREF1_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_AREF1;
        break;
    case TYPE_BYTE:
        inst = OP_AREF1B;
        break;
    case TYPE_STRING:
        inst = OP_AREF1S;
        break;
    default:
        Fatal("internal error: AREF1_inst");
        break;
    }
    return inst;
}

/* AREF2_inst - select a 2 index reference instruction */
static int AREF2_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_AREF2;
        break;
    case TYPE_BYTE:
        inst = OP_AREF2B;
        break;
    case TYPE_STRING:
        inst = OP_AREF2S;
        break;
    default:
        Fatal("internal error: AREF2_inst");
        break;
    }
    return inst;
}

/* ASET1_inst - select a 1 index set instruction */
static int ASET1_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_ASET1;
        break;
    case TYPE_BYTE:
        inst = OP_ASET1B;
        break;
    case TYPE_STRING:
        inst = OP_ASET1S;
        break;
    default:
        Fatal("internal error: ASET1_inst");
        break;
    }
    return inst;
}

/* ASET2_inst - select a 2 index set instruction */
static int ASET2_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_INTEGER:
    case TYPE_ARRAY:
        inst = OP_ASET2;
        break;
    case TYPE_BYTE:
        inst = OP_ASET2B;
        break;
    case TYPE_STRING:
        inst = OP_ASET2S;
        break;
    default:
        Fatal("internal error: ASET2_inst");
        break;
    }
    return inst;
}

/* codeaddr - get the current code address (actually, offset) */
int codeaddr(ParseContext *c)
{
    return (int)(c->cptr - c->codeBuf);
}

/* putcbyte - put a code byte into the code buffer */
int putcbyte(ParseContext *c, int b)
{
    int addr = codeaddr(c);
    if (c->cptr >= c->ctop)
        Fatal("Bytecode buffer overflow");
    *c->cptr++ = b;
    return addr;
}

/* putcword - put a code word into the code buffer */
int putcword(ParseContext *c, int w)
{
    int addr = codeaddr(c);
    if (c->cptr + 2 > c->ctop)
        Fatal("Bytecode buffer overflow");
    *c->cptr++ = w;
    *c->cptr++ = w >> 8;
    return addr;
}

/* merge - merge two reference chains */
int merge(ParseContext *c, int chn, int chn2)
{
    int last, nxt;

    /* if the chain we're adding is empty, just return the original chain */
    if (!chn2)
        return chn;

    /* find the last entry in the new chain */
    last = chn2;
    while (last != 0) {
        if (!(nxt = (int16_t)(c->codeBuf[last] | (c->codeBuf[last + 1] << 8))))
            break;
        last = nxt;
    }

    /* link the last entry in the new chain to the first entry in the original chain */
    c->codeBuf[last] = chn;
    c->codeBuf[last + 1] = chn >> 8;

    /* return the new chain now linked to the original chain */
    return chn2;
}

/* fixup - fixup a reference chain */
void fixup(ParseContext *c, int chn, int val)
{
    while (chn != 0) {
        int nxt = (int16_t)(c->codeBuf[chn] | (c->codeBuf[chn + 1] << 8));
        c->codeBuf[chn] = val;
        c->codeBuf[chn + 1] = val >> 8;
        chn = nxt;
    }
}

/* fixupbranch - fixup a reference chain */
void fixupbranch(ParseContext *c, int chn, int val)
{
    while (chn != 0) {
        int nxt = (int16_t)(c->codeBuf[chn] | (c->codeBuf[chn + 1] << 8));
        int off = val - (chn + 2); /* this assumes all three byte branch instructions */
        c->codeBuf[chn] = off;
        c->codeBuf[chn + 1] = off >> 8;
        chn = nxt;
    }
}
