/* db_symbols.c - symbol table routines
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "db_compiler.h"

/* InitSymbolTable - initialize a symbol table */
void InitSymbolTable(SymbolTable *table)
{
    table->head = NULL;
    table->pTail = &table->head;
}

/* FreeSymbols - free the symbols in a symbol table */
void FreeSymbols(ParseContext *c, SymbolTable *table)
{
    Symbol *sym, *next;
    for (sym = table->head; sym != NULL; sym = next) {
        next = sym->next;
        FreeType(c, sym->type);
        Free(sym);
    }
}

/* AddSymbol - add a symbol to a symbol table */
Symbol *AddSymbol(SymbolTable *table, char *name, StorageClass storageClass, Type *type, int offset, int value)
{
    size_t size = sizeof(Symbol) + strlen(name);
    Symbol *sym;
    
    /* allocate the symbol structure */
    sym = (Symbol *)Alloc("SYM", size);
    strcpy(sym->name, name);
    sym->storageClass = storageClass;
    sym->type = type;
    sym->offset = offset;
    sym->value = value;
    sym->next = NULL;

    /* add it to the symbol table */
    *table->pTail = sym;
    table->pTail = &sym->next;
    
    /* return the symbol */
    return sym;
}

/* FindSymbol - find a symbol in a symbol table */
Symbol *FindSymbol(SymbolTable *table, char *name)
{
    Symbol *sym = table->head;
    while (sym) {
        if (strcasecmp(name, sym->name) == 0)
            return sym;
        sym = sym->next;
    }
    return NULL;
}

/* DumpSymbols - dump a symbol table */
void DumpSymbols(SymbolTable *table, int showTags)
{
    int firstLine;
    Symbol *sym;
    for (sym = table->head, firstLine = VMTRUE; sym != NULL; sym = sym->next)
        if (sym->storageClass == SC_CONSTANT) {
            if (showTags && firstLine)
                printf("constants:\n");
            printf("  %s %04x\n", sym->name, sym->value);
            firstLine = VMFALSE;
        }
    for (sym = table->head, firstLine = VMTRUE; sym != NULL; sym = sym->next)
        if (sym->storageClass != SC_CONSTANT) {
            if (showTags && firstLine)
                printf("variables:\n");
            printf("  %04x: %s %04x\n", sym->offset, sym->name, sym->value);
            firstLine = VMFALSE;
        }
}
