///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2008
// 
// Filename: XGS_AVR_SOUND_DRV_V010.c
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 8.29.08
//
// Overview: The sound on the XGS AVR is output thru a single pin. The pin is connected thru a "low pass" RC filter, 
// which can handle either audible frequency digital waveforms OR PWM signals for full D/A conversion and reproduction
// of any sound. So for simple tones you can simply toggle the sound pin and you will hear the toggle rate, or you can
// output high frequency PWM and the RC filter will "average" or "integrate" the signal allowing for smooth D/A 
// waveform synthesis from 0..5V and frequencies that are well beyond the audible range with at least 8bits of 
// amplitude resolution and 8-12 bits of frequency resolution, give or take. This is feasbile since the hardware PWM
// unit can be used, thus to generate a PWM waveform, one must only update the PWM signal at a rate of the desired
// output frequency times the amplitude resolution, for example, if you wanted to synthesize a 1Khz PWM sine wave consisting
// of 8-bits of amplitude, that means 256 samples per cycle, thus the PWM would need to be updated at a min rate of:
// 256 * 1Khz = 256Khz, the PWM frequency itself should typically be 10x this rate, or 2.56 Mhz.
// 
// As noted above you can either manually toggle the pin or use the timer/counter functionality to do it for you. The 
// sound I/O pin is PD7 at Pin 16 of the AVR and is also the OC2A pin. Thus, for PWM or other counter output action you
// must control the timer/counter.
//
// I/O pin map for audio hardware
// 
// AUDIO_MONO | PD7 | Pin 16 | Output
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// this processor include needs to happen in the master file
// #define __AVR_ATmega644__

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/pgmspace.h>

// include local XGS API header files
#include "XGS_AVR_SYSTEM_V010.h"
#include "XGS_AVR_SOUND_DRV_V010.h"
//#include "XGS_AVR_UART_DRV_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MACROS /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Sound_Init(void)
{
// this function initializes the sound port I/O to output and does other minimal housekeeping since the sound
// functions haven't been developed, future additions to the init function should go here

// set data direction of sound output pin to output
SET_BIT(SND_DDR, SND_PIN); 

// set output to 0 for now
RESET_BIT(SND_PORT, SND_PIN); 

// return success, later add some more relevant error codes
return(1);

} // end Sound_Init

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Sound_Play(long freq)
{
// this function uses the internal timer/counter hardware to output a 50-50 duty cycle square wave on the sound pin
// the function returns immediately after called, so it can be used in real-time applications as a "poor mans" single
// channel audio system. The output is always a square wave though. Also, to keep things simple the function enables
// and initialized the counter hardware each call
// 
// freq - the frequency in hertz to output, 0 to turn the sound completely off and shutdown the timer/counter, the range
//        of valid frequencies is from 55 - 20,000Hz        

// char sbuffer[64];

// freq = 0 means turn off hardware
if (freq == 0)
    {
    // set clock pre-scaler and counter mode
    TCCR2B = (0 << CS22) | (0 << CS21)  | (0 << CS20);    
    } // end if
else
    {
    // set up timer for CTC mode, "clear timer on match", toggle mode, however, we need to select the correct
    // clock scaling bits to support the requested frequency so a little computation is needed to support the 
    // widest range of final frequencies based on the clock divider control bits:
    //    
    //    CSS22       CS21        CS20        Description
    //
    //    0           0           0           no clock source / timer stopped
    //    0           0           1           CLKin/1  
    //    0           1           0           CLKin/8  
    //    0           1           1           CLKin/32  
    //    1           0           0           CLKin/64
    //    1           0           1           CLKin/128
    //    1           1           0           CLKin/256
    //    1           1           1           CLKin/1024 <---- best choice for audio output
    // 
    // Note 1: The clock rates to the right are toggle rate, thus frequency is always 1/2 these
    // Note 2: The final CTC frequency is (CLKin / (2 * [1,8,32,64,128,256,1024] * (compare match + 1) )
    //
    // example: the system clock is 28_636_360 Hz, if we set the clock divider to 111 mode or divide by 1024 and the compare
    // match value to 255, the final output frequency will be: (28_636_360 hz / 2*(1024 * (255 + 1) )) = 54.6 Hz.
    // this example illustrates the basic challenge, since timer/counter 2 is only 8-bit we only can use the values
    // [0..255], thus the total range of frequencies available is rougly:
    //
    // OCR (compare match register) = 0, final output freq = (28_636_360 hz / 2*(1024 * (0 + 1) )) = 13982.59 Hz
    // OCR (compare match register) = 255, final output freq = (28_636_360 hz / 2*(1024 * (255 + 1) )) = 54.6 Hz
    //
    // Although, this seems to cover the entire range of frequencies, there is a problem... the accuracy of the final
    // desired frequency is 1/256th of the range, or in other words approx., (13982 - 54)/256 = 54 hz
    // additionally, there is another problem, the relation is an inverse relation, or log like, thus the accuracy
    // changes as the frequency changes, in other words for low values of OCR, the function is very inaccurate then for
    // larger values of OCR, it become very accurate and the min delta between values of frequency as a function of OCR
    // are realized, thus we need to think of the timer/counter as a set of curves, and we want to select the curve that
    // gives us the closest final output frequency to the frequency desired. A little math can show that the trick is to 
    // find the the smallest divider (1, 8, 32, ... 1024) that can satisfy our desired frequency, the smaller the divider,
    // the smaller the ranges deviance for each value of OCR as it increase. In any event, the final algorith is simple;
    // for any desired input frequency, we must solve for OCR, but we must also select the best divider constant to give 
    // us the most accurate output, it can be shown that if we pick a divider such that for the case of OCR=255, and then
    // plug it into the frequency formula, if our desired frequency is ABOVE this value then this is the divider we want,
    // so we will iterate thru a set of numbers, where each number represents a pre-computed fence value and then select 
    // the largest value that is smaller than our desired frequency
    // 
    // using the formula Fout =  (28_636_360 hz / 2*(N * (255 + 1) )), where N is the divider and equals 1,8,32,64,128,256,1024
    // we generate the following constants for the dividers N -> N:[1,8,32,64,128,256,1024] -> F[55930, 6991, 1747, 873, 436, 218, 54]

    // could use lookup tables, but just code it to save RAM and avoid a loop

    // equations for freq and OCR respectively
    // freq = (CLKin / (2 * N * (OCR+1) )
    // OCR = [CLKin / (1*N * freq)] - 1

    // first clip frequency, so its not lower than 55 Hz which is out of range for 8-bit counter and all possible divisors
    if ( freq < 55)
        freq = 55;

    // N=1 case
    if (freq  > 55930L)
        {
        // set output compare register value
        OCR2A  = (unsigned char) ( (long)f_cpu / ( (long)(1*2) * (long)freq ) - 1);
 
        // set clock pre-scaler to divide by 1 and counter mode to CTC mode, so whatever number we put into
        // the compare match register will count that many clocks per increment
        TCCR2B = (0 << WGM22) | (0 << CS22) | (0 << CS21)  | (1 << CS20);    

        // now re-compute actual frequency, so error can be determined
        freq = ( (long)f_cpu / ( (long)(2 * 1) * ( (long)OCR2A+1 ) ) );
        //sprintf(sbuffer, "\nOCR = %d, freq = %ld, div = 1", OCR2A, freq);
        //UART_Print_String(sbuffer);

        } // end if
    else
    // N=8 case
    if (freq  > 6991L)
        {
        // set output compare register value
        OCR2A  = (unsigned char) ( (long)f_cpu / ( (long)(8*2) * (long)freq ) - 1);

        // set clock pre-scaler to divide by 8 and counter mode to CTC mode, so whatever number we put into
        // the compare match register will count that many clocks per increment
        TCCR2B = (0 << WGM22) | (0 << CS22) | (1 << CS21)  | (0 << CS20);    

        // now re-compute actual frequency, so error can be determined
        freq = ((long)f_cpu / ((long)(2 * 8) * ((long)OCR2A+1) ) );
        //sprintf(sbuffer, "\nOCR = %d, freq = %ld, div = 8", OCR2A, freq);
        //UART_Print_String(sbuffer);

        } // end if
    else
    // N=32 case
    if (freq  > 1747L)
        {
        // set output compare register value
        OCR2A  = (unsigned char) ( (long)f_cpu / ( (long)(32*2) * (long)freq ) - 1);

        // set clock pre-scaler to divide by 32 and counter mode to CTC mode, so whatever number we put into
        // the compare match register will count that many clocks per increment
        TCCR2B = (0 << WGM22) | (0 << CS22) | (1 << CS21)  | (1 << CS20);    

        // now re-compute actual frequency, so error can be determined
        freq = ((long)f_cpu / ((long)(2 * 32) * ((long)OCR2A+1) ) );
        //sprintf(sbuffer, "\nOCR = %d, freq = %ld, div = 32", OCR2A, freq);
        //UART_Print_String(sbuffer);

        } // end if
    else
    // N=64 case
    if (freq  > 873L)
        {
        // set output compare register value
        OCR2A  = (unsigned char) ( (long)f_cpu / ( (long)(64*2) * (long)freq ) - 1);

        // set clock pre-scaler to divide by 64 and counter mode to CTC mode, so whatever number we put into
        // the compare match register will count that many clocks per increment
        TCCR2B = (0 << WGM22) | (1 << CS22) | (0 << CS21)  | (0 << CS20);    

        // now re-compute actual frequency, so error can be determined
        freq = ((long)f_cpu / ((long)(2 * 64) * ((long)OCR2A+1) ) );
        //sprintf(sbuffer, "\nOCR = %d, freq = %ld, div = 64", OCR2A, freq);
        //UART_Print_String(sbuffer);

        } // end if
    else
    // N=128 case
    if (freq  > 436L)
        {
        // set output compare register value
        OCR2A  = (unsigned char) ( (long)f_cpu / ( (long)(128*2) * (long)freq ) - 1);

        // set clock pre-scaler to divide by 128 and counter mode to CTC mode, so whatever number we put into
        // the compare match register will count that many clocks per increment
        TCCR2B = (0 << WGM22) | (1 << CS22) | (0 << CS21)  | (1 << CS20);    

        // now re-compute actual frequency, so error can be determined
        freq = ((long)f_cpu / ((long)(2 * 128) * ((long)OCR2A+1) ) );
        //sprintf(sbuffer, "\nOCR = %d, freq = %ld, div = 128", OCR2A, freq);
        //UART_Print_String(sbuffer);

        } // end if
    else
    // N=256 case
    if (freq  > 218L)
        {
        // set output compare register value
        OCR2A  = (unsigned char) ( (long)f_cpu / ( (long)(256*2) * (long)freq ) - 1);

        // set clock pre-scaler to divide by 2568 and counter mode to CTC mode, so whatever number we put into
        // the compare match register will count that many clocks per increment
        TCCR2B = (0 << WGM22) | (1 << CS22) | (1 << CS21)  | (0 << CS20);    

        // now re-compute actual frequency, so error can be determined
        freq = ((long)f_cpu / ((long)(2 * 256) * ((long)OCR2A+1) ) );
        //sprintf(sbuffer, "\nOCR = %d, freq = %ld, div = 256", OCR2A, freq);
        //UART_Print_String(sbuffer);

        } // end if
    else
    // N=1024 case,everthing else has to fit here
    if (freq > 54L)
        {
        // set output compare register value
        OCR2A  = (unsigned char) ( (long)f_cpu / ( (long)(1024*2) * (long)freq ) - 1);

        // set clock pre-scaler to divide by 1024 and counter mode to CTC mode, so whatever number we put into
        // the compare match register will count that many clocks per increment
        TCCR2B = (0 << WGM22) | (1 << CS22) | (1 << CS21)  | (1 << CS20);    

        // now re-compute actual frequency, so error can be determined
        freq = ((long)f_cpu / ((long)(2 * 1024) * ((long)OCR2A+1) ) );
        //sprintf(sbuffer, "\nOCR = %d, freq = %ld, div = 1024", OCR2A, freq);
        //UART_Print_String(sbuffer);

        } // end if


    // set control register for counter and waveform generator properly
    TCCR2A = (0 << COM2A1) | (1 << COM2A0) | (0 << COM2B1) | (0 << COM2B0) | (1 << WGM21) | (0 << WGM20);

    } // end else

// return actual frequency that timer/counter was able to be set to, error usually 0-5%
return( (int)freq );

} // end Sound_Play

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Sound_Play_Toggle(long freq, long time)
{
// this is a very simple function that simply toggles the sound output pin for a certain amount of time and at a certain 
// rate, while the function is working it blocks all other processing, any interrupts will continue to run, but distort
// the sound. This function is for simple testing only and making little beeps and blips!
// 
// freq - frequency of square wave to output (1hz to 12Khz all valid)
// time - time to play sound in millisecond, eg. 1000 would be 1 second since 1000 ms = 1 second.

long counter, cycles;
float scale;

// compute total number of times to run waveform loop, note this become innaccurate as the frequency increases, due to
// the overhead of the function calls executing so much, but as said, this is just for crude sounds, scale adjusts this
scale = (((float)100000-((float)freq*(float)freq/(float)2000))/(float)100000);

cycles = (long)(scale * (float)freq*((float)time/(float)1000));

// generate the sound for the proper time
for (counter = 0; counter < cycles; counter++)
    {
    SET_BIT(SND_PORT, SND_PIN);
    Delay_Clocks( (f_cpu/(2*freq)) );
    RESET_BIT(SND_PORT, SND_PIN);
    Delay_Clocks( (f_cpu/(2*freq)) );
    } // end for

} // end Sound_Play_Toggle
