/* db_vmfcnpic.c - intrinsic functions
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdlib.h>
#include "../runtime/db_vm.h"

#include "XGS_PIC_GFX_DRV_V010.h"
#include "XGS_PIC_GAMEPAD_DRV_V010.h"
#include "XGS_PIC_SOUND_DRV_V010.h"
#include "XGS_PIC_NTSC_TILE2_V010.h"
#include "XGS_PIC_NTSC_160_192_2_V010.h"
#include "db_text.h"

/* graphics state */
static int graphicsStarted;

/* bitmap graphics variables */
unsigned char *g_VRAMBuffer;
unsigned short *g_PaletteMap;
unsigned char *g_Palettes;

/* tile graphics variables */
static unsigned char *myTiles;
static int myTileCount;
static unsigned char *myTileMap;
static unsigned char *myColorTable;
static Sprite *mySpriteTable;
static Sprite mySprites[MAX_SPRITES];

static void InitBitmapGraphics(Interpreter *i);
static void TryInitTileGraphics(Interpreter *i);

void VM_fcninit(void)
{
    /* graphics not started yet */
    graphicsStarted = VMFALSE;

    /* initialize the bitmap graphics variables */
    g_VRAMBuffer = NULL;
    g_PaletteMap = NULL;
    g_Palettes = NULL;

    /* initialize the tile graphics variables */
    myTiles = NULL;
    myTileCount = 0;
    myTileMap = NULL;
    myColorTable = NULL;
    mySpriteTable = NULL;
}

void VM_fcnstop(void)
{
    GFX_StopDrawing();
}

void fcn_waitforvsync(Interpreter *i)
{
	if (g_Drawing) {
        WAIT_FOR_VSYNC_END();	
	    WAIT_FOR_VSYNC_START();	
	}
    Drop(i, 1);
}

void fcn_gamepad(Interpreter *i)
{
	int which = i->sp[0];
	i->sp[-1] = Gamepad_Read(which == 0 ? GAMEPAD_0 : GAMEPAD_1);
	Drop(i, 1);
}

void fcn_sound(Interpreter *i)
{
	int freq = i->sp[0];
	SND_PlayTone(freq);
	Drop(i, 2);
}

static void InitBitmapGraphics(Interpreter *i)
{
    int sizeNeeded, freeSize;
    uint8_t *free;

    sizeNeeded = VRAM_BUF_SIZE + PALETTE_MAP_SIZE * sizeof(unsigned short) + MIN_PALETTE_SIZE;

    if (!(free = VM_FreeSpace(&freeSize)) || freeSize < sizeNeeded)
        Abort(i, "insufficient memory");

    g_VRAMBuffer = (unsigned char *)free;
    g_PaletteMap = (unsigned short *)(g_VRAMBuffer + VRAM_BUF_SIZE);
    g_Palettes = (unsigned char *)(g_PaletteMap + PALETTE_MAP_SIZE);

    memset(g_VRAMBuffer, 0, VRAM_BUF_SIZE);
    memset(g_PaletteMap, 0, PALETTE_MAP_SIZE * sizeof(unsigned short));
    
    g_Palettes[0] = NTSC_BLACK;
    g_Palettes[1] = NTSC_RED;
    g_Palettes[2] = NTSC_GREEN;
    g_Palettes[3] = NTSC_BLUE;

    GFX_InitBitmap(BITMAP_SCREEN_WIDTH, BITMAP_SCREEN_HEIGHT, BITMAP_SCREEN_BPP, g_VRAMBuffer, g_Palettes, g_PaletteMap);
    GFX_StartDrawing(SCREEN_TYPE_NTSC);
}

void fcn_plot(Interpreter *i)
{
	int x = i->sp[-2];
	int y = i->sp[-1];
	int color = i->sp[0];
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
	GFX_Plot_2BPP(x, y, color, g_VRAMBuffer);
	Drop(i, 4);
}

void fcn_line(Interpreter *i)
{
	int x1 = i->sp[-4];
	int y1 = i->sp[-3];
	int x2 = i->sp[-2];
	int y2 = i->sp[-1];
	int color = i->sp[0];
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
	GFX_Line_2BPP(x1, y1, x2, y2, color, g_VRAMBuffer);
	Drop(i, 6);
}

void fcn_hline(Interpreter *i)
{
	int x1 = i->sp[-3];
	int x2 = i->sp[-2];
	int y = i->sp[-1];
	int color = i->sp[0];
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
	GFX_HLine_2BPP(x1, x2, y, color, g_VRAMBuffer);
	Drop(i, 5);
}

void fcn_vline(Interpreter *i)
{
	int x = i->sp[-3];
	int y1 = i->sp[-2];
	int y2 = i->sp[-1];
	int color = i->sp[0];
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
	GFX_VLine_2BPP(x, y1, y2, color, g_VRAMBuffer);
	Drop(i, 5);
}

void fcn_clear(Interpreter *i)
{
	int color = i->sp[0];
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
    GFX_FillScreen_2BPP(color, g_VRAMBuffer);
	Drop(i, 2);
}

void fcn_fillrect(Interpreter *i)
{
	int x = i->sp[-4];
	int y = i->sp[-3];
	int w = i->sp[-2];
	int h = i->sp[-1];
	int color = i->sp[0];
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
	GFX_FillRegion_2BPP(x, y, w, h, color, g_VRAMBuffer);
	Drop(i, 6);
}

void fcn_circle(Interpreter *i)
{
	int x = i->sp[-3];
	int y = i->sp[-2];
	int r = i->sp[-1];
	int color = i->sp[0];
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
	GFX_Circle_2BPP(x, y, r, color, g_VRAMBuffer);
	Drop(i, 5);
}

void fcn_position(Interpreter *i)
{
    int x = i->sp[-1];
    int y = i->sp[0];
	VM_displaypos(x, y);
	Drop(i, 3);
}

void fcn_text(Interpreter *i)
{
    int x = i->sp[-2];
    int y = i->sp[-1];
	int inHeap;
    char *str = GetStringPtr(i, (uint16_t)i->sp[0], &inHeap);
    int len = GetStringLen(i, i->sp[0]);
    if (!g_VRAMBuffer) InitBitmapGraphics(i);
    while (--len >= 0) {
        DrawBitmapFontChar(VMGETBYTE(str++, inHeap), x, y, g_VRAMBuffer);
        x += 6;
    }
    StringRelease(i, i->sp[0]);
	Drop(i, 4);
}

/* start graphics when all of the tables have been setup */
static void TryInitTileGraphics(Interpreter *i)
{
    if (!mySpriteTable) {
        int i;
        for (i = 0; i < MAX_SPRITES; ++i)
            mySprites[i].State = 0;
        mySpriteTable = mySprites;
    }
    if (myTiles && myTileMap && myColorTable) {
        GFX_InitTile(myTiles, myTileMap, myColorTable, mySpriteTable);
        if (!graphicsStarted) {
            GFX_StartDrawing(SCREEN_TYPE_NTSC);
            graphicsStarted = VMTRUE;
        }
    }
}

void fcn_settilemap(Interpreter *i)
{
	int16_t *array = &i->variables[i->sp[0]];
    int16_t dim1 = array[-2];
    int16_t dim2 = array[-1] / dim1;
    if (dim1 >= SCREEN_TILE_MAP_WIDTH && dim2 >= SCREEN_TILE_MAP_HEIGHT) {
        myTileMap = (unsigned char *)array;
        TryInitTileGraphics(i);
    }
	Drop(i, 2);
}

void fcn_settiles(Interpreter *i)
{
	int16_t *array = &i->variables[i->sp[0]];
    int16_t dim1 = array[-2];
    int16_t dim2 = array[-1] / dim1;
    int nTiles = dim2 / TILE_HEIGHT;
    if (dim1 == TILE_WIDTH && nTiles > 0 && (dim2 % TILE_HEIGHT) == 0) {
        myTiles = (unsigned char *)array;
        myTileCount = nTiles;
        TryInitTileGraphics(i);
    }
	Drop(i, 2);
}

void fcn_setcolortable(Interpreter *i)
{
	int16_t *array = &i->variables[i->sp[0]];
    int16_t dim1 = array[-2];
    int16_t dim2 = array[-1] / dim1;
    if (dim1 == MAX_PALETTE_SIZE && dim2 == 1) {
        myColorTable = (unsigned char *)array;
        TryInitTileGraphics(i);
    }
	Drop(i, 2);
}

void fcn_setsprite(Interpreter *i)
{
    int16_t n = i->sp[-4];
    int16_t state = i->sp[-3];
    int16_t x = i->sp[-2];
    int16_t y = i->sp[-1];
    int16_t tile = i->sp[0];
    if (n >= 0 && n < MAX_SPRITES && myTiles && tile >= 0 && tile < myTileCount) {
        Sprite *sprite = &mySprites[n];
        sprite->State = state;
        sprite->X = x;
        sprite->Y = y;
        sprite->PixelData = &myTiles[tile];
        TryInitTileGraphics(i);
    }
    Drop(i, 6);
}

void VM_displayinit(void)
{
    InitBitmapGraphics(NULL);
}

#define CHARW   6
#define CHARH   8

static int xStart = 0;
static int xPosition = 0;
static int yPosition = 0;

void VM_displaypos(int x, int y)
{
    xStart = xPosition = x;
    yPosition = y;
}

void VM_displaychar(int ch)
{
    if (ch == '\n') {
        xPosition = xStart;
        yPosition += CHARH;
    }
    else {
        if (!g_VRAMBuffer) InitBitmapGraphics(&interp);
        DrawBitmapFontChar(ch, xPosition, yPosition, g_VRAMBuffer);
        xPosition += CHARW;
    }
}

void VM_displaystr(char *str)
{
    while (*str)
        VM_displaychar(*str++);
}

void VM_displaytab(void)
{
    do {
        VM_displaychar(' ');
    } while (((xPosition - xStart) % (CHARW * 8)) != 0);
}

